import unittest
from os import environ
from sqlalchemy import create_engine
from sqlalchemy.orm import scoped_session, sessionmaker
from sqlalchemy.ext.declarative import declarative_base
from flask_security import SQLAlchemySessionUserDatastore
from database.models import User, Role, Application
from database.database import init_db


class DatabaseSetupTest(unittest.TestCase):

    def setUp(self):
        # Setup local database for testing
        self.engine = create_engine('sqlite://', connect_args={'check_same_thread': False})
        init_db(self.engine)
        session_factory = sessionmaker(bind=self.engine)
        Session = scoped_session(session_factory)
        self.new_session = Session()
        self.user_datastore = SQLAlchemySessionUserDatastore(
            self.new_session, User, Role)

    def tearDown(self):
        pass

    def test_table_availability(self):
        self.assertTrue(self.engine.has_table('oas_user'))
        self.assertTrue(self.engine.has_table('oas_application'))
        self.assertTrue(self.engine.has_table('oas_role'))

    def test_connect_to_database_with_provided_config(self):
        db_user = environ["DATABASE_USER"]
        db_password = environ["DATABASE_PASSWORD"]
        db_host = environ["DATABASE_HOST"]
        db_name = environ["DATABASE_NAME"]
        engine = create_engine('postgresql://{0}:{1}@{2}/{3}'.format(
            db_user, db_password, db_host, db_name))
        self.assertTrue(engine.connect())

    def test_create_and_activate_user_with_role(self):
        user = self.user_datastore.create_user(
            email="admin@example.net",
            username="admin",
            password="password")
        self.user_datastore.activate_user(user)
        self.assertTrue(user.is_active)
        admin_role = self.user_datastore.create_role(name="admin", description="Admins")
        self.assertIsInstance(admin_role, Role)
        self.user_datastore.add_role_to_user(user, admin_role)
        self.assertTrue(user.has_role(admin_role))

    def test_create_an_application_and_grant_user_access(self):
        Session = sessionmaker(bind=self.engine)
        session = Session()
        user = self.user_datastore.create_user(
            email="admin@example.net",
            username="admin",
            password="password")
        self.user_datastore.commit()
        application = Application(
            name="nextcloud",
            description='''
                This application can be used to do something like a or b. After
                reading this description the user should know what to expect
                from the application''')
        session.add(application)
        nextcloud = session.query(Application).filter_by(name="nextcloud").first()
        self.assertIs(application, nextcloud)
        admin_user = session.query(User).filter_by(username="admin").first()
        self.assertEqual(admin_user.id, user.id)
        nextcloud.users.append(admin_user)
        self.assertIn(nextcloud, admin_user.applications)

